// src/index.ts
import { addNamed } from "@babel/helper-module-imports";
import {
  arrayPattern,
  arrowFunctionExpression,
  assignmentExpression,
  binaryExpression,
  blockStatement,
  callExpression,
  expressionStatement,
  identifier,
  isArrayPattern,
  isArrowFunctionExpression,
  isAssignmentExpression,
  isAssignmentPattern,
  isBlockStatement,
  isExpressionStatement,
  isFunctionDeclaration,
  isFunctionExpression,
  isIdentifier,
  isLabeledStatement,
  isMemberExpression,
  isObjectMethod,
  isObjectProperty,
  isVariableDeclaration,
  isVariableDeclarator,
  numericLiteral,
  objectExpression,
  objectProperty,
  returnStatement,
  stringLiteral,
  variableDeclaration,
  variableDeclarator
} from "@babel/types";
function getHookIdentifier(hooks, path, name) {
  const current = hooks.get(name);
  if (current) {
    return current;
  }
  const newID = addNamed(path, name, "solid-js");
  hooks.set(name, newID);
  return newID;
}
function signalSingleExpression(hooks, path, signalIdentifier, stateIdentifier) {
  const readIdentifier = path.scope.generateUidIdentifier(signalIdentifier.name);
  const writeIdentifier = path.scope.generateUidIdentifier(`set${signalIdentifier.name}`);
  const expr = variableDeclaration("const", [variableDeclarator(arrayPattern([
    readIdentifier,
    writeIdentifier
  ]), callExpression(getHookIdentifier(hooks, path, "createSignal"), [
    stateIdentifier,
    objectExpression([
      objectProperty(identifier("name"), stringLiteral(signalIdentifier.name))
    ])
  ]))]);
  path.insertAfter(expr);
  const parent = path.scope.path;
  if (parent) {
    parent.traverse({
      ObjectProperty(p) {
        if (p.scope !== path.scope && p.scope.hasOwnBinding(signalIdentifier.name)) {
          return;
        }
        if (p.node.shorthand && isIdentifier(p.node.key) && p.node.key.name === signalIdentifier.name && isIdentifier(p.node.value) && p.node.value.name === signalIdentifier.name) {
          p.replaceWith(objectProperty(signalIdentifier, callExpression(readIdentifier, [])));
        }
      },
      Identifier(p) {
        if (p.node.name !== signalIdentifier.name) {
          return;
        }
        if (p.scope !== path.scope && p.scope.hasOwnBinding(signalIdentifier.name)) {
          return;
        }
        if (isObjectMethod(p.parent) && p.parent.key === p.node) {
          return;
        }
        if (isObjectProperty(p.parent) && p.parent.key === p.node) {
          return;
        }
        if (isVariableDeclarator(p.parent) && p.parent.id === p.node) {
          return;
        }
        if (isArrayPattern(p.parent) && p.parent.elements.includes(p.node)) {
          return;
        }
        if (isArrowFunctionExpression(p.parent) && p.parent.params.includes(p.node)) {
          return;
        }
        if (isFunctionExpression(p.parent) && p.parent.params.includes(p.node)) {
          return;
        }
        if (isFunctionDeclaration(p.parent) && p.parent.params.includes(p.node)) {
          return;
        }
        if (isLabeledStatement(p.parent) && p.parent.label === p.node) {
          return;
        }
        if (isMemberExpression(p.parent) && p.parent.property === p.node) {
          return;
        }
        if (isFunctionDeclaration(p.parent) && p.parent.id === p.node) {
          return;
        }
        if (isAssignmentPattern(p.parent) && p.parent.left === p.node && (isArrowFunctionExpression(p.parentPath.parent) && p.parentPath.parent.params.includes(p.parent) || isFunctionDeclaration(p.parentPath.parent) && p.parentPath.parent.params.includes(p.parent) || isFunctionExpression(p.parentPath.parent) && p.parentPath.parent.params.includes(p.parent))) {
          return;
        }
        p.replaceWith(callExpression(readIdentifier, []));
      },
      UpdateExpression(p) {
        if (p.scope !== path.scope && p.scope.hasOwnBinding(signalIdentifier.name)) {
          return;
        }
        if (isIdentifier(p.node.argument) && p.node.argument.name === signalIdentifier.name) {
          const param = p.scope.generateUidIdentifier("current");
          if (p.node.prefix) {
            p.replaceWith(callExpression(writeIdentifier, [
              arrowFunctionExpression([param], binaryExpression(p.node.operator === "++" ? "+" : "-", param, numericLiteral(1)))
            ]));
          } else {
            p.replaceWith(callExpression(arrowFunctionExpression([], blockStatement([
              variableDeclaration("const", [
                variableDeclarator(param, callExpression(readIdentifier, []))
              ]),
              expressionStatement(callExpression(writeIdentifier, [
                arrowFunctionExpression([], binaryExpression(p.node.operator === "++" ? "+" : "-", param, numericLiteral(1)))
              ])),
              returnStatement(param)
            ])), []));
          }
        }
      },
      AssignmentExpression(p) {
        if (p.scope !== path.scope && p.scope.hasOwnBinding(signalIdentifier.name)) {
          return;
        }
        const identifier2 = p.node.left;
        const expression = p.node.right;
        if (isIdentifier(identifier2) && identifier2.name === signalIdentifier.name) {
          if (p.node.operator === "=") {
            p.replaceWith(callExpression(writeIdentifier, [
              arrowFunctionExpression([], expression)
            ]));
          } else {
            const param = p.scope.generateUidIdentifier("current");
            p.replaceWith(callExpression(writeIdentifier, [
              arrowFunctionExpression([param], assignmentExpression(p.node.operator, param, expression))
            ]));
          }
        }
      }
    });
  }
}
function signalExpression(hooks, path, body) {
  if (isExpressionStatement(body)) {
    if (!isAssignmentExpression(body.expression)) {
      throw new Error("Expected assignment expression");
    }
    if (body.expression.operator !== "=") {
      throw new Error("Invalid assignment expression operator");
    }
    const leftExpr = body.expression.left;
    const rightExpr = body.expression.right;
    if (!isIdentifier(leftExpr)) {
      throw new Error("Expected identifier");
    }
    signalSingleExpression(hooks, path, leftExpr, rightExpr);
    path.remove();
  } else if (isVariableDeclaration(body)) {
    for (let i = body.declarations.length - 1; i >= 0; i -= 1) {
      const declarator = body.declarations[i];
      const leftExpr = declarator.id;
      const rightExpr = declarator.init;
      if (!isIdentifier(leftExpr)) {
        throw new Error("Expected identifier");
      }
      signalSingleExpression(hooks, path, leftExpr, rightExpr != null ? rightExpr : identifier("undefined"));
    }
    path.remove();
  } else {
    throw new Error("Expected assignment expression or variable declaration");
  }
}
function memoSingleExpression(hooks, path, memoIdentifier, stateIdentifier, replace = false) {
  const readIdentifier = path.scope.generateUidIdentifier(memoIdentifier.name);
  const expr = variableDeclaration("const", [variableDeclarator(readIdentifier, callExpression(getHookIdentifier(hooks, path, "createMemo"), [
    arrowFunctionExpression([], stateIdentifier),
    objectExpression([
      objectProperty(identifier("name"), stringLiteral(memoIdentifier.name))
    ])
  ]))]);
  if (replace) {
    path.replaceWith(expr);
  } else {
    path.insertAfter(expr);
  }
  const parent = path.scope.path;
  if (parent) {
    parent.traverse({
      ObjectProperty(p) {
        if (p.scope !== path.scope && p.scope.hasOwnBinding(memoIdentifier.name)) {
          return;
        }
        if (p.node.shorthand && isIdentifier(p.node.key) && p.node.key.name === memoIdentifier.name && isIdentifier(p.node.value) && p.node.value.name === memoIdentifier.name) {
          p.replaceWith(objectProperty(memoIdentifier, callExpression(readIdentifier, [])));
        }
      },
      Identifier(p) {
        if (p.node.name !== memoIdentifier.name) {
          return;
        }
        if (p.scope !== path.scope && p.scope.hasOwnBinding(memoIdentifier.name)) {
          return;
        }
        if (isObjectMethod(p.parent) && p.parent.key === p.node) {
          return;
        }
        if (isObjectProperty(p.parent) && p.parent.key === p.node) {
          return;
        }
        if (isVariableDeclarator(p.parent) && p.parent.id === p.node) {
          return;
        }
        if (isArrayPattern(p.parent) && p.parent.elements.includes(p.node)) {
          return;
        }
        if (isArrowFunctionExpression(p.parent) && p.parent.params.includes(p.node)) {
          return;
        }
        if (isFunctionExpression(p.parent) && p.parent.params.includes(p.node)) {
          return;
        }
        if (isFunctionDeclaration(p.parent) && p.parent.params.includes(p.node)) {
          return;
        }
        if (isLabeledStatement(p.parent) && p.parent.label === p.node) {
          return;
        }
        if (isMemberExpression(p.parent) && p.parent.property === p.node) {
          return;
        }
        if (isFunctionDeclaration(p.parent) && p.parent.id === p.node) {
          return;
        }
        if (isAssignmentPattern(p.parent) && p.parent.left === p.node && (isArrowFunctionExpression(p.parentPath.parent) && p.parentPath.parent.params.includes(p.parent) || isFunctionDeclaration(p.parentPath.parent) && p.parentPath.parent.params.includes(p.parent) || isFunctionExpression(p.parentPath.parent) && p.parentPath.parent.params.includes(p.parent))) {
          return;
        }
        p.replaceWith(callExpression(readIdentifier, []));
      }
    });
  }
}
function memoExpression(hooks, path, body) {
  if (isExpressionStatement(body)) {
    if (isAssignmentExpression(body.expression)) {
      if (body.expression.operator !== "=") {
        throw new Error("Invalid assignment expression operator");
      }
      const leftExpr = body.expression.left;
      const rightExpr = body.expression.right;
      if (!isIdentifier(leftExpr)) {
        throw new Error("Expected identifier");
      }
      memoSingleExpression(hooks, path, leftExpr, rightExpr, true);
    } else {
      throw new Error("Expected assignment expression");
    }
  } else if (isVariableDeclaration(body)) {
    for (let i = body.declarations.length - 1; i >= 0; i -= 1) {
      const declarator = body.declarations[i];
      const leftExpr = declarator.id;
      const rightExpr = declarator.init;
      if (!isIdentifier(leftExpr)) {
        throw new Error("Expected identifier");
      }
      memoSingleExpression(hooks, path, leftExpr, rightExpr != null ? rightExpr : identifier("undefined"));
    }
    path.remove();
  } else {
    throw new Error("Expected assignment expression or variable declaration");
  }
}
function effectExpression(hooks, path, body) {
  let callback;
  if (isBlockStatement(body)) {
    callback = arrowFunctionExpression([], body);
  } else if (isExpressionStatement(body)) {
    callback = body.expression;
  } else {
    throw new Error("Expected arrow function or block expression");
  }
  path.replaceWith(callExpression(getHookIdentifier(hooks, path, "createEffect"), [
    callback
  ]));
}
function computedExpression(hooks, path, body) {
  let callback;
  if (isBlockStatement(body)) {
    callback = arrowFunctionExpression([], body);
  } else if (isExpressionStatement(body)) {
    callback = body.expression;
  } else {
    throw new Error("Expected arrow function or block expression");
  }
  path.replaceWith(callExpression(getHookIdentifier(hooks, path, "createComputed"), [
    callback
  ]));
}
function renderEffectExpression(hooks, path, body) {
  let callback;
  if (isBlockStatement(body)) {
    callback = arrowFunctionExpression([], body);
  } else if (isExpressionStatement(body)) {
    callback = body.expression;
  } else {
    throw new Error("Expected arrow function or block expression");
  }
  path.replaceWith(callExpression(getHookIdentifier(hooks, path, "createRenderEffect"), [
    callback
  ]));
}
function mountExpression(hooks, path, body) {
  let callback;
  if (isBlockStatement(body)) {
    callback = arrowFunctionExpression([], body);
  } else if (isExpressionStatement(body)) {
    callback = body.expression;
  } else {
    throw new Error("Expected arrow function or block expression");
  }
  path.replaceWith(callExpression(getHookIdentifier(hooks, path, "onMount"), [
    callback
  ]));
}
function cleanupExpression(hooks, path, body) {
  let callback;
  if (isBlockStatement(body)) {
    callback = arrowFunctionExpression([], body);
  } else if (isExpressionStatement(body)) {
    callback = body.expression;
  } else {
    throw new Error("Expected arrow function or block expression");
  }
  path.replaceWith(callExpression(getHookIdentifier(hooks, path, "onCleanup"), [
    callback
  ]));
}
function errorExpression(hooks, path, body) {
  let callback;
  if (isBlockStatement(body)) {
    callback = arrowFunctionExpression([], body);
  } else if (isExpressionStatement(body)) {
    callback = body.expression;
  } else {
    throw new Error("Expected arrow function or block expression");
  }
  path.replaceWith(callExpression(getHookIdentifier(hooks, path, "onError"), [
    callback
  ]));
}
function rootExpression(hooks, path, body) {
  let callback;
  if (isBlockStatement(body)) {
    callback = arrowFunctionExpression([], body);
  } else if (isExpressionStatement(body)) {
    callback = body.expression;
  } else {
    throw new Error("Expected arrow function or block expression");
  }
  path.replaceWith(callExpression(getHookIdentifier(hooks, path, "createRoot"), [
    callback
  ]));
}
function untrackExpression(hooks, path, body) {
  let callback;
  if (isBlockStatement(body)) {
    callback = arrowFunctionExpression([], body);
  } else if (isExpressionStatement(body)) {
    callback = body.expression;
  } else {
    throw new Error("Expected arrow function or block expression");
  }
  path.replaceWith(callExpression(getHookIdentifier(hooks, path, "untrack"), [
    callback
  ]));
}
function batchExpression(hooks, path, body) {
  let callback;
  if (isBlockStatement(body)) {
    callback = arrowFunctionExpression([], body);
  } else if (isExpressionStatement(body)) {
    callback = body.expression;
  } else {
    throw new Error("Expected arrow function or block expression");
  }
  path.replaceWith(callExpression(getHookIdentifier(hooks, path, "batch"), [
    callback
  ]));
}
function solidReactivityPlugin() {
  return {
    pre() {
      this.hooks = new Map();
    },
    visitor: {
      LabeledStatement(path, state) {
        const { label, body } = path.node;
        switch (label.name) {
          case "signal":
            signalExpression(state.hooks, path, body);
            break;
          case "effect":
            effectExpression(state.hooks, path, body);
            break;
          case "computed":
            computedExpression(state.hooks, path, body);
            break;
          case "renderEffect":
            renderEffectExpression(state.hooks, path, body);
            break;
          case "memo":
            memoExpression(state.hooks, path, body);
            break;
          case "mount":
            mountExpression(state.hooks, path, body);
            break;
          case "cleanup":
            cleanupExpression(state.hooks, path, body);
            break;
          case "error":
            errorExpression(state.hooks, path, body);
            break;
          case "untrack":
            untrackExpression(state.hooks, path, body);
            break;
          case "batch":
            batchExpression(state.hooks, path, body);
            break;
          case "root":
            rootExpression(state.hooks, path, body);
            break;
          default:
            break;
        }
      }
    }
  };
}
export {
  solidReactivityPlugin as default
};
//# sourceMappingURL=index.js.map
